import datetime
from typing import Iterable
from rotafy.config import chore


class NoPersonName(Exception):
    def __init__(self) -> None:
        super().__init__("Must provide a person name in the TOML configuration file.")


class PersonNotFound(Exception):
    def __init__(self, person_name: str) -> None:
        super().__init__(
            f"Cannot find person named {person_name} among the list of people."
        )


class Person:
    def __init__(
        self,
        name: str,
        skills: Iterable[chore.Chore],
        telephone: str = "",
        unavailable: Iterable[datetime.date] = [],
        training: Iterable[chore.Chore] = [],
    ) -> None:
        self.name = name
        self.telephone = telephone
        self.skills = set(skills)
        self.unavailable = set(unavailable)

        self._raw_training = set(training)
        self.experience = {c: 0 for c in self._raw_training if c not in self.skills}

    def __repr__(self) -> str:
        init_args = (
            self.name,
            self.skills,
            self.telephone,
            self.unavailable,
            self._raw_training,
        )
        reprs = (repr(arg) for arg in init_args)
        s = f"Person({', '.join(reprs)})"
        return s

    def __str__(self) -> str:
        return self.name

    def __eq__(self, other) -> bool:
        return other and self.name == other.name and self.telephone == other.telephone

    def __hash__(self) -> int:
        return hash((self.name, self.telephone))

    def qualified(self, chore: chore.Chore) -> bool:
        return chore in self.skills

    def is_learning(self, chore: chore.Chore) -> bool:
        return chore in self.experience.keys()

    def add_to_experience(self, chore: chore.Chore) -> None:
        if chore in self.skills:
            return

        if chore in self.experience.keys():
            self.experience[chore] += 1
        else:
            self.experience[chore] = 1

        qualification_threshold = (
            chore.num_training_sessions + chore.num_shadowing_sessions
        )
        if self.experience[chore] >= qualification_threshold:
            del self.experience[chore]
            self.skills.add(chore)

    def reduce_experience(self, chore: chore.Chore) -> None:
        if chore in self.skills:
            qualification_threshold = (
                chore.num_training_sessions + chore.num_shadowing_sessions
            )
            self.experience[chore] = qualification_threshold
            self.skills = set(s for s in self.skills if s != chore)

        if chore not in self.experience.keys():
            self.experience[chore] = 0

        self.experience[chore] = max(0, self.experience[chore] - 1)

    def is_shadowing(self, chore: chore.Chore) -> bool:
        if chore not in self.experience.keys():
            return False

        if self.experience[chore] < chore.num_training_sessions:
            return True

        return False

    def is_being_observed(self, chore: chore.Chore) -> bool:
        if chore not in self.experience.keys():
            return False

        if self.experience[chore] >= chore.num_training_sessions:
            return True

        return False

    def available(self, date: datetime.date) -> bool:
        return not (date in self.unavailable)

    def can_do(self, chore: chore.Chore, date: datetime.date) -> bool:
        return self.qualified(chore) and self.available(date)

    def can_be_trained(self, chore: chore.Chore, date: datetime.date) -> bool:
        return self.is_learning(chore) and self.available(date)

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        if len(value.strip()) == 0:
            raise NoPersonName

        self._name = value.strip()


def find_person(person_name: str, people: Iterable[Person]) -> Person:
    for person in people:
        if person.name == person_name:
            return person

    for person in people:
        if person.name.lower() == person_name.lower():
            return person

    raise PersonNotFound(person_name)



import datetime
from rotafy.config import chore, person


class NotQualified(Exception):
    def __init__(self, person: person.Person, chore: chore.Chore) -> None:
        super().__init__(f"{person.name} is not qualified to do {chore.name}.")


class PersonUnavailable(Exception):
    def __init__(self, person: person.Person, date: datetime.date) -> None:
        super().__init__(f"{person.name} is not available on {date}.")


class ChoreNotScheduled(Exception):
    def __init__(self, chore: chore.Chore, date: datetime.date) -> None:
        super().__init__(f"{chore} is not scheduled to happen on {date}.")


class Assignment:

    def __init__(
        self,
        date: datetime.date,
        chore: chore.Chore,
        person: person.Person,
        trainee: person.Person | None = None,
        notification_sent: bool = False,
    ) -> None:

        if person.qualified(chore) == False:
            raise NotQualified(person, chore)

        if person.available(date) == False:
            raise PersonUnavailable(person, date)

        if trainee is not None and trainee.available(date) == False:
            raise PersonUnavailable(person, date)

        if chore.on(date) == False:
            raise ChoreNotScheduled(chore, date)

        if person == trainee:
            trainee = None

        self.date = date
        self.chore = chore
        self.person = person
        self.trainee = trainee
        self.notification_sent = notification_sent

    def __repr__(self):
        init_args = (
            self.date,
            self.chore,
            self.person,
            self.trainee,
            self.notification_sent,
        )
        reprs = (repr(arg) for arg in init_args)
        s = f"Assignment({', '.join(reprs)})"
        return s

    def __str__(self):
        if self.trainee is None:
            return self.person.name

        if self.trainee.is_shadowing(self.chore):
            return f"{self.person.name} with {self.trainee.name} shadowing"

        if self.trainee.is_being_observed(self.chore):
            return f"{self.trainee.name} supervised by {self.person.name}"

        return f"{self.person.name} with {self.trainee.name}"

    def __eq__(self, other) -> bool:
        return (
            other
            and self.date == other.date
            and self.chore == other.chore
            and self.person == other.person
            and self.trainee == other.trainee
        )

    def __hash__(self) -> int:
        return hash(
            (
                self.date,
                self.chore,
                self.person,
                self.trainee if self.trainee is not None else None,
            )
        )

    def mark_notified(self) -> None:
        self.notification_sent = True



import toml
from typing import Iterable
from rotafy.config import chore, person


class Config:
    def __init__(self, toml_file_path: str) -> None:
        self.path = toml_file_path
        self.raw = toml.load(self.path)

        self.name = self.raw["name"]

        self.lookahead_days = self.raw.get("lookahead_days", 14)
        self.clicksend_username = self.raw.get("clicksend_username", "")
        self.clicksend_api_key = self.raw.get("clicksend_api_key", "")
        self.message_template = self.raw.get(
            "message",
            "Hi {{recipient}}! On {{date}}, {{chore}} is due to be handled by "
            "{{assignment}}. Thanks!",
        )

        self.chores = set()
        for ordinal, raw_chore in enumerate(self.raw.get("chore")):
            this_chore_notify = raw_chore.get("notify", False)
            if isinstance(this_chore_notify, bool) and this_chore_notify == True:
                this_chore_notify = self.raw.get("default_notification_days", 1)

            this_chore = chore.Chore(
                raw_chore.get("name"),
                ordinal,
                raw_chore.get("recurrence"),
                this_chore_notify,
                raw_chore.get(
                    "required_training_sessions",
                    self.raw.get("default_number_of_training_sessions", 1),
                ),
                raw_chore.get(
                    "required_shadowing_sessions",
                    self.raw.get("default_number_of_shadowing_sessions", 1),
                ),
                raw_chore.get("exceptions", []),
            )
            self.chores.add(this_chore)

        self.people = set()
        for raw_person in self.raw.get("person"):
            this_person_skills = self._get_chores_from_names(
                raw_person.get("skills", [])
            )
            this_person_training = self._get_chores_from_names(
                raw_person.get("training", [])
            )

            this_person = person.Person(
                raw_person.get("name"),
                this_person_skills,
                str(raw_person.get("telephone", "")),
                raw_person.get("unavailable", []),
                this_person_training,
            )
            self.people.add(this_person)

    def __repr__(self):
        return f"Config({repr(self.path)})"

    def __str__(self):
        s = self.name + "\n"
        s += "  Chores:\n"
        for c in self.chores:
            s += "    - " + str(c) + "\n"

        s += "  People:\n"
        for p in self.people:
            s += "    - " + str(p) + "\n"

        return s.strip()

    def __eq__(self, other) -> bool:
        return other and self.name == other.name

    def _get_chores_from_names(self, names: Iterable[str]) -> Iterable[chore.Chore]:
        if len(names) == 1:
            singleton_name = names[0].lower()
            if singleton_name in ("any", "all"):
                return self.chores

        found_chores = set()
        for name in names:
            found_chores.add(chore.find_chore(name, self.chores))

        return found_chores



import click
import logging
from rotafy.api import manager


@click.group()
@click.argument(
    "configuration_file",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True),
    required=True,
)
@click.option(
    "--verbose", "-v", is_flag=True, default=False, help="Print verbose log messages."
)
@click.pass_context
def cli(ctx, configuration_file, verbose):
    log_format = "%(levelname)s @ %(asctime)s - %(message)s"
    log_level = logging.ERROR
    if verbose:
        log_level = logging.INFO

    logging.basicConfig(format=log_format, level=log_level)

    ctx.obj = manager.Manager(configuration_file)


@cli.command("print", help="Print the upcoming rota to the screen.")
@click.pass_obj
def print_to_screen(m):
    m.print()


@cli.command(help="Output the upcoming rota to a PDF file.")
@click.argument("filename", type=click.Path(exists=False), required=True)
@click.pass_obj
def to_pdf(m, filename):
    m.to_pdf(filename)


@cli.command(help="Send notifications to individuals with upcoming chores.")
@click.pass_obj
def notify(m):
    m.notify()


@cli.command(help="Assign a person to a chore on a given date.")
@click.argument("date", type=click.DateTime(), required=True)
@click.argument("chore", type=click.STRING, required=True)
@click.argument("person", type=click.STRING, required=True)
@click.pass_obj
def assign(m, date, chore, person):
    m.add_person(date.date(), chore, person)


@cli.command(help="Swap two peoples' assigned chores on the same date.")
@click.argument("date", type=click.DateTime(), required=True)
@click.argument("person1", type=click.STRING, required=True)
@click.argument("person2", type=click.STRING, required=True)
@click.pass_obj
def swap(m, date, person1, person2):
    m.swap(date.date(), person1, person2)


@cli.command(help="Replace a person with another on a given date.")
@click.argument("date", type=click.DateTime(), required=True)
@click.argument("person", type=click.STRING, required=True)
@click.argument("replacement", type=click.STRING, required=True)
@click.pass_obj
def replace(m, date, person, replacement):
    m.replace(date.date(), person, replacement)


if __name__ == "__main__":
    cli()



import datetime
from collections import Counter
from typing import Iterable
from rotafy.config import chore, person
from rotafy.rota import assignment


class NoAssignments(Exception):
    def __init__(self) -> None:
        super().__init__("Need to provide one or more chore assignments.")


class MultipleDates(Exception):
    def __init__(self, dates: Iterable[datetime.date]) -> None:
        super().__init__(
            f"Chore assignments must be all on the same date. Given dates: {dates}."
        )


class ChoreAssignedMultipleTimes(Exception):
    def __init__(self, chore: chore.Chore) -> None:
        super().__init__(f"Cannot assign {chore.name} more than once on the same date.")


class PersonAssignedMultipleTimes(Exception):
    def __init__(self, person: person.Person) -> None:
        super().__init__(
            f"Cannot assign {person.name} more than once on the same date."
        )


class Row:
    def __init__(self, assignments: Iterable[assignment.Assignment]) -> None:
        self.assignments = assignments
        self.date = list(self.assignments)[0].date

    def __getitem__(self, chore: chore.Chore) -> assignment.Assignment | None:
        matches = [a for a in self.assignments if a.chore == chore]
        if len(matches) != 1:
            return None

        return matches[0]

    def __setitem__(
        self, chore: chore.Chore, new_assignment: assignment.Assignment
    ) -> None:
        kept_assignments = [a for a in self.assignments if a.chore != chore]
        override = kept_assignments + [new_assignment]
        self.assignments = override

    def __delitem__(self, chore: chore.Chore) -> None:
        kept_assignments = [a for a in self.assignments if a.chore != chore]
        self.assignments = kept_assignments

    @property
    def assignments(self) -> Iterable[assignment.Assignment]:
        return self._assignments

    @assignments.setter
    def assignments(self, assignments: Iterable[assignment.Assignment]) -> None:
        num_assignments = len(assignments)

        if num_assignments == 0:
            raise NoAssignments

        distinct_dates = set(a.date for a in assignments)
        if len(distinct_dates) > 1:
            raise MultipleDates(distinct_dates)

        chore_counts = Counter(a.chore for a in assignments)
        if any(v > 1 for v in chore_counts.values()):
            raise ChoreAssignedMultipleTimes(chore_counts.most_common(1)[0][0])

        people_counts = Counter(a.person for a in assignments)
        people_counts.update(a.trainee for a in assignments if a.trainee is not None)
        if any(v > 1 for v in people_counts.values()):
            raise PersonAssignedMultipleTimes(people_counts.most_common(1)[0][0])

        self._assignments = assignments
        self._assignments.sort(key=lambda a: a.chore.ordinal)



import clicksend_client
import logging
import datetime
import jinja2
import ast
from rotafy.config import person
from rotafy.rota import assignment, printable


logger = logging.getLogger(__name__)


class APIStatusNotSuccessful(Exception):
    def __init__(self, status_message: str) -> None:
        super().__init__(
            f"API returned status message '{status_message}'. See https://developers.clicksend.com/docs/#status-codes for more information."
        )


class Notifier:
    def __init__(
        self, clicksend_username: str, clicksend_api_key: str, message_template: str
    ) -> None:
        clicksend_config = clicksend_client.Configuration()
        clicksend_config.username = clicksend_username
        clicksend_config.password = clicksend_api_key
        configured_client = clicksend_client.ApiClient(clicksend_config)
        self.clicksend_api = clicksend_client.SMSApi(configured_client)

        jinja_env = jinja2.Environment(loader=jinja2.BaseLoader())
        self.template = jinja_env.from_string(message_template)

        self.queue = []

    def format_upcoming_date(self, date: datetime.date) -> str:
        date_ordinal = printable.ordinal(date.day)

        today = datetime.date.today()
        days_to_date = (date - today).days
        if days_to_date < 7:
            day_of_week = date.strftime("%A")
            return f"{day_of_week} ({date_ordinal})"

        return printable.human_readable_date(date, True, False)

    def add_to_queue(
        self, recipient: person.Person, assignment_to_notify: assignment.Assignment
    ) -> None:
        assignment_str = str(assignment_to_notify)
        assignment_msg = assignment_str.replace(recipient.name, "you")

        message = self.template.render(
            recipient=recipient.name,
            date=self.format_upcoming_date(assignment_to_notify.date),
            chore=assignment_to_notify.chore.name,
            assignment=assignment_msg,
        )

        logger.info(f"Adding message '{message}' to {recipient.telephone} to queue")
        sms = clicksend_client.SmsMessage(
            source="Rotafy", body=message, to=recipient.telephone
        )
        self.queue.append(sms)

    def message_from_assignment(
        self, assignment_to_notify: assignment.Assignment
    ) -> None:
        self.add_to_queue(assignment_to_notify.person, assignment_to_notify)
        if assignment_to_notify.trainee is not None:
            self.add_to_queue(assignment_to_notify.trainee, assignment_to_notify)

    def send(self) -> None:
        messages_to_send = clicksend_client.SmsMessageCollection(messages=self.queue)
        try:
            api_response = self.clicksend_api.sms_send_post(messages_to_send)
        except Exception as e:
            raise e
        else:
            logger.info(f"API Response: {api_response}")

            api_response_data = ast.literal_eval(api_response)
            statuses = [m["status"] for m in api_response_data["data"]["messages"]]
            unsuccessful_statuses = [
                status for status in statuses if status != "SUCCESS"
            ]
            if len(unsuccessful_statuses) > 0:
                raise APIStatusNotSuccessful(unsuccessful_statuses[0])

            logger.info(f"All {len(self.queue)} messages in queue sent")
            self.queue = []



import datetime
import logging
import pkg_resources
import os
import pickle
import copy
from typing import Iterable
from rotafy.rota import row


logger = logging.getLogger(__name__)


class MismatchedDates(Exception):
    def __init__(self, set_date: datetime.date, new_date: datetime.date) -> None:
        super().__init__(
            f"Date of new row ({new_date}) must match the and date being set ({set_date})."
        )


class Rota:
    def __init__(self, name: str) -> None:
        self.name = name
        self.file_path = pkg_resources.resource_filename(
            __name__, f"/rotas/{self.name}.pkl"
        )
        self.rows = []
        self.load()
        self.sort()

    def __getitem__(self, date: datetime.date) -> row.Row | None:
        matches = [r for r in self.rows if r.date == date]
        if len(matches) != 1:
            return None

        return matches[0]

    def __setitem__(self, date: datetime.date, new_row: row.Row) -> None:
        if date != new_row.date:
            raise MismatchedDates(date, new_row.date)

        return self.add_row(new_row)

    def __delitem__(self, date: datetime.date) -> None:
        return self.delete_row(date)

    def load(self) -> None:
        if os.path.exists(self.file_path):
            with open(self.file_path, "rb") as f:
                self.rows = pickle.load(f)

    def save(self) -> None:
        if os.path.exists(self.file_path) == False:
            os.makedirs(os.path.dirname(self.file_path), exist_ok=True)

        with open(self.file_path, "wb") as f:
            pickle.dump(self.rows, f)

    def sort(self) -> None:
        self.rows.sort(key=lambda r: r.date)

    def add_row(self, new_row: row.Row) -> None:
        new_row_s = [f"{a.chore.name}: {str(a)}" for a in new_row.assignments]
        logger.info(f"Adding row {', '.join(new_row_s)} to {new_row.date}")

        all_row_dates = set(r.date for r in self.rows)
        if new_row.date in all_row_dates:
            self.delete_row(new_row.date)

        self.rows.append(copy.deepcopy(new_row))
        self.sort()

    def delete_row(self, date: datetime.date) -> None:
        logger.info(f"Deleting row from {date}")
        self.rows = [row for row in self.rows if row.date != date]

    def rows_prior(self, date: datetime.date, inc: bool = False) -> Iterable[row.Row]:
        self.sort()

        if inc:
            return [row for row in self.rows if row.date <= date]

        return [row for row in self.rows if row.date < date]

    def rows_after(self, date: datetime.date, inc: bool = False) -> Iterable[row.Row]:
        self.sort()

        if inc:
            return [row for row in self.rows if row.date >= date]

        return [row for row in self.rows if row.date > date]

    @property
    def latest_date(self) -> datetime.date:
        if len(self.rows) == 0:
            return datetime.date.today()

        return max(row.date for row in self.rows)



import datetime
import pandas
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from rotafy.rota import rota


class PrintableRota(rota.Rota):
    def __init__(self, name: str) -> None:
        super().__init__(name)

    def __str__(self) -> str:
        return self.dataframe.to_string()

    def _draw_table_figure(self) -> matplotlib.figure.Figure:
        df_separate = self.dataframe.copy()
        width = len(df_separate.columns)
        height = df_separate.shape[0]
        if height == 0:
            return None

        heading_colour = (0.083, 0.203, 0.273)  # primary blue

        plt.rcParams["font.family"] = "Inter,sans-serif"
        plt.rcParams["font.size"] = 11
        fig, ax = plt.subplots()
        ax.axis("tight")
        ax.axis("off")

        table = ax.table(
            cellText=df_separate.values,
            cellLoc="center",
            rowLabels=df_separate.index,
            rowLoc="right",
            rowColours=[heading_colour] * height,
            colLabels=df_separate.columns,
            colColours=[heading_colour] * width,
            colLoc="center",
            loc="center",
        )

        for c in range(width):
            table[0, c].get_text().set_color("white")

        for r in range(height):
            table[r + 1, -1].get_text().set_color("white")

        return fig

    def pdf(self, output_file: str) -> None:
        fig = self._draw_table_figure()
        if fig is not None:
            with PdfPages(output_file) as pdf:
                pdf.savefig(fig, bbox_inches="tight")

            plt.close()

    def print(self) -> None:
        print(self.__str__())

    @property
    def dataframe(self) -> pandas.DataFrame:
        all_chores = set(a.chore for r in self.rows for a in r.assignments)
        ordered_chores = list(all_chores)
        ordered_chores.sort(key=lambda c: c.ordinal)
        ordered_chore_names = [chore.name for chore in ordered_chores]

        self.sort()

        data = {}
        for r in self.rows:
            row_data = []
            for c in ordered_chores:
                a = r[c]
                if a is None:
                    row_data.append(None)
                else:
                    row_data.append(str(a))

            data[r.date] = row_data

        df = pandas.DataFrame.from_dict(
            data, orient="index", columns=ordered_chore_names
        )
        today = datetime.date.today()
        min_ts = pandas.Timestamp(today).date()
        df = df[df.index >= min_ts]
        df.index = df.index.map(human_readable_date)
        df.fillna("-", inplace=True)
        return df


def ordinal(n: int) -> str:
    return f"{n:d}{'tsnrhtdd'[(n//10%10!=1)*(n%10<4)*n%10::4]}"


def human_readable_date(
    date: datetime.date,
    include_day_of_week: bool = False,
    include_year: bool = True,
) -> str:
    date_ordinal = ordinal(date.day)

    format = f"{date_ordinal} %B"
    if include_day_of_week:
        format = "%A " + format

    if include_year:
        format += " %Y"

    human_readable = date.strftime(format)
    return human_readable



import datetime
import logging
import itertools
from typing import Iterable
import random
from retry.api import retry_call
from clicksend_client.rest import ApiException
from rotafy.config import config, chore, person
from rotafy.rota import printable, assignment, row
from rotafy.api import notifier


logger = logging.getLogger(__name__)


class DateNotFound(Exception):
    def __init__(self, date: datetime.date) -> None:
        super().__init__(f"No existing rota row on {date}. Try `add_person` instead.")


class PersonNotAssigned(Exception):
    def __init__(self, date: datetime.date, person_name: str) -> None:
        super().__init__(
            f"{person_name} is not assigned to a chore on {date}. Try `add_person` instead."
        )


class ReplacementPersonAlreadyAssigned(Exception):
    def __init__(self, date: datetime.date, person_name: str) -> None:
        super().__init__(
            f"{person_name} is already assigned to a chore on {date}. Try `swap` instead."
        )


class NoValidAssignments(Exception):
    def __init__(self, date: datetime.date) -> None:
        super().__init__(f"Cannot find any valid assignments on {date}.")


class Manager:
    def __init__(self, toml_file_path: str) -> None:
        self.configuration = config.Config(toml_file_path)
        self.name = self.configuration.name

        logger.info(f"Creating rotafy.Manager named {self.name}")
        logger.info(f"Loaded configuration file from {toml_file_path}")

        self.rota = printable.PrintableRota(self.name)
        self.notifier = notifier.Notifier(
            self.configuration.clicksend_username,
            self.configuration.clicksend_api_key,
            self.configuration.message_template,
        )
        self.update_experience()
        self.check_and_heal()

    def print(self) -> None:
        self.rota.print()

    def to_pdf(self, output_file: str) -> None:
        self.rota.pdf(output_file)

    def chores_on(self, date: datetime.date) -> Iterable[chore.Chore]:
        found_chores = set(c for c in self.configuration.chores if c.on(date))
        logger.info(f"Found chores on {date}: {[c.name for c in found_chores]}")
        return found_chores

    def update_experience(self):
        for r in self.rota.rows:
            for a in r.assignments:
                if a.trainee is not None:
                    t = person.find_person(a.trainee.name, self.configuration.people)
                    t.add_to_experience(a.chore)

    def find_assignment(
        self, date: datetime.date, person_name: str
    ) -> None | assignment.Assignment:
        existing_row = self.rota[date]
        if existing_row is None:
            logger.info(f"No existing row on {date}")
            return None

        person_assignment = [
            a
            for a in existing_row.assignments
            if a.person.name == person_name or a.trainee.name == person_name
        ]
        if len(person_assignment) != 1:
            logger.info(f"No existing assignment for {person_name} on {date}")
            return None

        logger.info(f"Found existing assignment for {person_name} on {date}")
        return person_assignment[0]

    def remove_person(self, date: datetime.date, person_name: str) -> None:
        existing_row = self.rota[date]
        if existing_row is None:
            raise DateNotFound(date)

        person_assignment = self.find_assignment(date, person_name)
        if person_assignment is None:
            raise PersonNotAssigned(date, person_name)

        logger.info(
            f"Removing {person_name} from {person_assignment.chore.name} on {date}"
        )

        new_row = existing_row
        if person_assignment.trainee.name == person_name:
            person_assignment.trainee.reduce_experience(person_assignment.chore)
            person_assignment.trainee = None
            new_row[person_assignment.chore] = person_assignment
        else:
            del new_row[person_assignment.chore]

        if len(new_row.assignments) == 0:
            del self.rota[date]
        else:
            self.rota[date] = new_row

        self.rota.save()

    def add_person(
        self, date: datetime.date, chore_name: str, person_name: str
    ) -> None:
        chore_to_do = chore.find_chore(chore_name, self.configuration.chores)
        person_to_assign = person.find_person(person_name, self.configuration.people)
        new_assignment = assignment.Assignment(date, chore_to_do, person_to_assign)

        logger.info(f"Adding {person_name} to {chore_name} on {date}")

        existing_row = self.rota[date]
        if existing_row is None:
            new_row = row.Row([new_assignment])
        else:
            new_row = existing_row
            new_row[chore_to_do] = new_assignment

        self.rota[date] = new_row
        self.rota.save()

    def swap(self, date: datetime.date, person1_name: str, person2_name: str) -> None:
        existing_row = self.rota[date]
        if existing_row is None:
            raise DateNotFound(date)

        person1_assignment = self.find_assignment(date, person1_name)
        if person1_assignment is None:
            raise PersonNotAssigned(date, person1_name)

        person2_assignment = self.find_assignment(date, person2_name)
        if person2_assignment is None:
            raise PersonNotAssigned(date, person2_name)

        logger.info(f"Swapping {person1_name} and {person2_name} on {date}")

        person1_assigned_as_trainee = person1_assignment.trainee.name == person1_name
        person2_assigned_as_trainee = person2_assignment.trainee.name == person2_name

        self.remove_person(date, person1_name)
        self.remove_person(date, person2_name)

        if not (person1_assigned_as_trainee):
            self.add_person(date, person1_assignment.chore, person2_name)

        if not (person2_assigned_as_trainee):
            self.add_person(date, person2_assignment.chore, person1_name)

    def replace(
        self, date: datetime.date, person_name: str, replacement_name: str
    ) -> None:
        replacement_assignment = self.find_assignment(date, replacement_name)
        if replacement_assignment is not None:
            raise ReplacementPersonAlreadyAssigned(date, replacement_name)

        logger.info(f"Replacing {person_name} with {replacement_name} on {date}")

        self.remove_person(date, person_name)
        self.add_person(date, replacement_name)

    def check_and_heal(self):
        # Remove all chores that no longer should be carried out and assigned people
        # that are no longer available.
        upcoming_rota = self.rota.rows_after(datetime.date.today(), True)
        for row in upcoming_rota:
            for a in row.assignments:
                try:
                    updated_chore = chore.find_chore(
                        a.chore.name, self.configuration.chores
                    )
                except chore.ChoreNotFound:
                    logger.info(
                        f"Removing {a.chore.name} from {a.date} - no longer configured"
                    )
                    if len([_ for _ in row.assignments if _.chore != a.chore]) == 0:
                        del self.rota[row.date]
                    else:
                        del self.rota[row.date][a.chore]

                try:
                    updated_person = person.find_person(
                        a.person.name, self.configuration.people
                    )
                except person.PersonNotFound:
                    logger.info(
                        f"Removing {a.person.name} from {a.date} - no longer configured"
                    )
                    if len([_ for _ in row.assignments if _.chore != a.chore]) == 0:
                        del self.rota[row.date]
                    else:
                        del self.rota[row.date][a.chore]

                updated_trainee = None
                if a.trainee is not None:
                    try:
                        updated_trainee = person.find_person(
                            a.trainee.name, self.configuration.people
                        )
                    except person.PersonNotFound:
                        logger.info(
                            f"Removing {a.trainee.name} as trainee from {a.date} - no longer configured"
                        )
                        a.trainee.reduce_experience(a.chore)
                        a.trainee = None
                        self.rota[row.date][a.chore] = a
                        updated_trainee = None

                if (
                    updated_chore.on(a.date) == False
                    or updated_person.available(a.date) == False
                ):
                    if updated_chore.on(a.date) == False:
                        logger.info(f"Removing {a.chore.name} - no longer on {a.date}")
                    elif updated_person.available(a.date) == False:
                        logger.info(
                            f"Removing {a.person.name} - no longer available on {a.date}"
                        )

                    if len([_ for _ in row.assignments if _.chore != a.chore]) == 0:
                        del self.rota[row.date]
                    else:
                        del self.rota[row.date][a.chore]
                else:
                    if (
                        updated_trainee is not None
                        and updated_trainee.available(a.date) == False
                    ):
                        logger.info(
                            f"Removing {a.trainee.name} - no longer available on {a.date}"
                        )
                        a.trainee.reduce_experience(a.chore)
                        a.trainee = None
                        self.rota[row.date][a.chore] = a

        self.fill()

    def all_independently_valid_assignments(
        self, date: datetime.date, chore_to_assign: chore.Chore
    ) -> Iterable[assignment.Assignment]:
        valid_people = [
            p for p in self.configuration.people if p.can_do(chore_to_assign, date)
        ]
        valid_trainees = [None] + [
            t
            for t in self.configuration.people
            if t.can_be_trained(chore_to_assign, date)
        ]
        valid_assignments = []
        for p, t in itertools.product(valid_people, valid_trainees):
            new_assignment = assignment.Assignment(date, chore_to_assign, p, t)
            valid_assignments.append(new_assignment)

        logging.info(
            f"Found valid assignments for {chore_to_assign.name} on {date}: {[str(a) for a in valid_assignments]}"
        )
        return valid_assignments

    def row_weight(self, row_to_check: row.Row) -> float:
        date = row_to_check.date
        previous_rows = self.rota.rows_prior(date)
        previous_rows.reverse()
        max_look_back = 10
        previous_rows = [r for i, r in enumerate(previous_rows) if i < max_look_back]

        # For each person, working from most recent row to least recent row, we
        # will subtract the following from the weight.
        # latest:        1/2 if chore matches, 1/4 if person is assigned any chore
        # second latest: 1/4 if chore matches, 1/8 if person is assigned any chore
        # etc.
        # Therefore the most this weight could ever be reduced by is
        #   1/2 + 1/4 + 1/8 + ... + 1/(2 ^ max_look_back)
        # = ((2 ^ max_look_back) - 1) / 2 ^ max_look_back
        # = very close to but never > 1
        weight = len(self.configuration.chores)

        for assignment_to_check in row_to_check.assignments:
            n = 0
            for comparison_row in previous_rows:
                same_chore_assignment = [
                    a
                    for a in comparison_row.assignments
                    if a.chore == assignment_to_check.chore
                ]
                # Only increment n if the chore is actually being done on this
                # date.
                if len(same_chore_assignment) == 0:
                    continue

                n += 1
                if assignment_to_check.person == same_chore_assignment[0].person:
                    weight -= 1 / (2**n)
                else:
                    not_same_chore = [
                        a
                        for a in comparison_row.assignments
                        if a.chore != assignment_to_check.chore
                    ]
                    for a in not_same_chore:
                        # Otherwise, subtract a smaller value if the person has been
                        # assigned any chore on that date.
                        if assignment_to_check.person == a.person:
                            weight -= 1 / ((2**n) * 2)
                            break

        row_s = [f"{a.chore.name}: {str(a)}" for a in row_to_check.assignments]
        logger.info(f"Weight of {weight} calculated for {', '.join(row_s)}")
        return weight

    def assign_chores_on(self, date: datetime.date) -> None:
        chores_on_date = self.chores_on(date)
        if len(chores_on_date) == 0:
            logger.info(f"No chores on {date} to assign")
            return

        existing_row = self.rota[date]
        existing_assignments = []
        if existing_row is not None:
            existing_assignments = [existing_row[c] for c in chores_on_date]

        existing_assignments = [a for a in existing_assignments if a is not None]
        existing_chores = [a.chore for a in existing_assignments]
        logger.info(f"Existing chores already assigned on {date}: {existing_chores}")

        chores_to_assign = [c for c in chores_on_date if c not in existing_chores]
        logger.info(f"New assignments required on {date} for {chores_to_assign}")
        if len(chores_to_assign) == 0:
            return

        # TODO: Work from the chore with the least possible assignments to the most.
        for c in chores_to_assign:
            existing_row = self.rota[date]
            existing_assignments = []
            if existing_row is not None:
                existing_assignments = [existing_row[c] for c in chores_on_date]
            existing_assignments = [a for a in existing_assignments if a is not None]
            unchanged_assignments = [a for a in existing_assignments if a.chore != c]

            choices = self.all_independently_valid_assignments(date, c)
            valid_rows = []
            for choice in choices:
                try:
                    new_row = row.Row(unchanged_assignments + [choice])
                except Exception:
                    pass
                else:
                    valid_rows.append(new_row)

            if len(valid_rows) == 0:
                logger.info(f"No valid combination of assignments found for {date}")
                if len(existing_assignments) > 0:
                    logger.info(f"Re-evaluating all chores due on {date}")
                    del self.rota[date]
                    self.assign_chores_on(date)
                else:
                    raise NoValidAssignments(date)

            weights = [self.row_weight(r) for r in valid_rows]
            max_weight = max(weights)
            index_with_max = [i for i, w in enumerate(weights) if w == max_weight]
            best_rows = [valid_rows[i] for i in index_with_max]
            best_row = random.choice(best_rows)

            for a in best_row.assignments:
                if a.trainee is not None:
                    a.trainee.add_to_experience(c)

            self.rota.add_row(best_row)

    def fill(self) -> None:
        today = datetime.date.today()
        all_lookahead_days = [
            today + datetime.timedelta(days=days_to_add)
            for days_to_add in range(self.configuration.lookahead_days + 1)
        ]
        for r in self.rota.rows_after(today, True):
            if r.date not in all_lookahead_days:
                all_lookahead_days.append(row.date)

        chores_on_dates = [
            date for date in all_lookahead_days if len(self.chores_on(date)) > 0
        ]
        logger.info(f"Attempting to fill assignments for {chores_on_dates}")
        chores_on_dates.sort()
        for date in chores_on_dates:
            self.assign_chores_on(date)

        self.rota.save()

    def notify(self) -> None:
        today = datetime.date.today()

        for c in self.configuration.chores:
            if c.notify == False:
                continue

            notification_cut_off = today + datetime.timedelta(days=c.notify)
            for r in self.rota.rows_prior(notification_cut_off, True):
                a = r[c]
                if a is not None and a.notification_sent == False:
                    self.notifier.message_from_assignment(a)
                    try:
                        retry_call(
                            _notifier_send,
                            fargs=[self.notifier],
                            exceptions=ApiException,
                            tries=3,
                            delay=5,
                            backoff=5,
                        )
                    except Exception as e:
                        raise e
                    else:
                        a.mark_notified()
                        self.rota.save()


def _notifier_send(notifier: notifier.Notifier) -> None:
    notifier.send()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rotafy', 'rotafy.api', 'rotafy.config', 'rotafy.rota']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.7,<9.0.0',
 'clicksend-client>=5.0.78,<6.0.0',
 'jinja2>=3.1.4,<4.0.0',
 'matplotlib>=3.9.0,<4.0.0',
 'pandas>=2.2.2,<3.0.0',
 'python-dateutil>=2.9.0.post0,<3.0.0',
 'recurrent>=0.4.1,<0.5.0',
 'retry>=0.9.2,<0.10.0',
 'toml>=0.10.2,<0.11.0']

entry_points = \
{'console_scripts': ['rotafy = rotafy.cli:cli']}

setup_kwargs = {
    'name': 'rotafy',
    'version': '0.1.0',
    'description': 'Build, update, and delegate a rota/roster of chores, notifying individuals when their scheduled task is upcoming',
    'long_description': '# rotafy\nBuild, update, and delegate a rota/roster of chores, notifying individuals when their scheduled task is upcoming\n\n## Contributing\n\nRun `pre-commit install`.',
    'author': 'Ryan McKeown',
    'author_email': 'ryanmckeown@mail4me.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from recurrent.event_parser import RecurringEvent
import datetime
from dateutil import rrule
from typing import Iterable


class NoChoreName(Exception):
    def __init__(self) -> None:
        super().__init__("Must provide a chore name in the TOML configuration file.")


class ChoreNotFound(Exception):
    def __init__(self, chore_name: str) -> None:
        super().__init__(
            f"Cannot find chore named {chore_name} among the list of chores."
        )


class Chore:
    def __init__(
        self,
        name: str,
        ordinal: int,
        recurrence: str,
        notify: int | bool,
        num_training_sessions: int,
        num_shadowing_sessions: int,
        exceptions: Iterable[datetime.date] = [],
    ) -> None:
        self.name = name
        self.ordinal = ordinal
        self._raw_recurrence = recurrence
        self.recurring_rule = generate_rrule(recurrence)
        self.notify = notify
        self.num_training_sessions = num_training_sessions
        self.num_shadowing_sessions = num_shadowing_sessions
        self.exceptions = exceptions

    def __repr__(self) -> str:
        init_args = (
            self.name,
            self.ordinal,
            self._raw_recurrence,
            self.notify,
            self.num_training_sessions,
            self.num_shadowing_sessions,
            self.exceptions,
        )
        reprs = (repr(arg) for arg in init_args)
        s = f"Chore({', '.join(reprs)})"
        return s

    def __str__(self) -> str:
        return self.name

    def __eq__(self, other) -> bool:
        return other and self.name == other.name

    def __hash__(self) -> int:
        return hash(self.name)

    def on(self, date: datetime.date) -> bool:
        timestamp = datetime.datetime.combine(date, datetime.time.min)
        rrule_date = self.recurring_rule.after(timestamp, inc=True).date()
        in_rrule = rrule_date == date
        if date in self.exceptions:
            return not (in_rrule)

        return in_rrule

    def next(self, start_date: datetime.date) -> datetime.date:
        timestamp = datetime.datetime.combine(start_date, datetime.time.min)
        next_date = self.recurring_rule.after(timestamp)
        while next_date.date() in self.exceptions:
            next_date = self.recurring_rule.after(next_date)

        return next_date.date()

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        if len(value.strip()) == 0:
            raise NoChoreName

        self._name = value.strip()


def generate_rrule(recurrence: str) -> rrule.rrule:
    start_of_today = datetime.datetime.combine(datetime.date.today(), datetime.time.min)
    recurring_event = RecurringEvent(now_date=start_of_today)
    recurring_event_rrule = recurring_event.parse(recurrence.lower())
    rule = rrule.rrulestr(recurring_event_rrule)
    if isinstance(rule, rrule.rrule):
        rule = rule.replace(dtstart=start_of_today)

    if isinstance(rule, rrule.rruleset):
        rule._rrule = [r.replace(dtstart=start_of_today) for r in rule._rrule]

    return rule


def find_chore(chore_name: str, chores: Iterable[Chore]) -> Chore:
    for chore in chores:
        if chore.name == chore_name:
            return chore

    for chore in chores:
        if chore.name.lower() == chore_name.lower():
            return chore

    raise ChoreNotFound(chore_name)
